from dataclasses import dataclass


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


canada = Country(
    languages=["English", "French"],
    capital="Ottawa",
    population=39_000_000,
    citizens=[
        Citizen(
            name="Olivier",
            birthyear=1985,
            hometown="Montreal",
        ),
        Citizen(
            name="Abraham",
            birthyear=2018,
            hometown="Shawinigan",
        ),
    ],
)


world = World(countries={"canada": canada})


big_world = World(
    countries={
        f"country_{i}": Country(
            languages=[f"Language_{i}_A", f"Language_{i}_B"],
            capital=f"Capital_{i}",
            population=1_000_000 + i * 50_000,
            citizens=[
                Citizen(
                    name=f"Citizen_{i}_{j}",
                    birthyear=1970 + (j % 50),
                    hometown=f"Hometown_{i}_{j}",
                )
                for j in range(100)
            ],
        )
        for i in range(100)
    }
)

roboland = Country(
    languages=[f"Robolang{i}" for i in range(10000)],
    capital="Robopolis",
    population=1000,
    citizens=[
        Citizen(
            f"Bobot{i}",
            birthyear=3000 + i,
            hometown=f"Bobotown{i}",
        )
        for i in range(1000)
    ],
)



from typing import Literal

import pytest
from ovld import Medley

from serieux import Serieux, dump, load, serialize
from serieux.ctx import Context, ModifyContext, OmitDefaults, Trail
from serieux.exc import ValidationError

from .common import has_312_features, one_test_per_assert
from .definitions import (
    Color,
    Defaults,
    DIDHolder,
    DotDict,
    File,
    Level,
    LTHolder,
    Point,
    Thingies,
)


@one_test_per_assert
def test_serialize_scalars():
    assert serialize(0) == 0
    assert serialize(12) == 12
    assert serialize(-3.25) == -3.25
    assert serialize("flagada") == "flagada"
    assert serialize(True) is True
    assert serialize(False) is False
    assert serialize(None) is None


@one_test_per_assert
def test_serialize_scalars_conversion():
    assert serialize(float, 10) == 10.0


def test_serialize_point():
    pt = Point(1, 2)
    assert serialize(Point, pt) == {"x": 1, "y": 2}


def test_serialize_list_of_points():
    pts = [Point(1, 2), Point(3, 4)]
    assert serialize(list[Point], pts) == [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
    ]


def test_serialize_dict_of_points():
    pts = {"p1": Point(1, 2), "p2": Point(3, 4)}
    assert serialize(dict[str, Point], pts) == {
        "p1": {"x": 1, "y": 2},
        "p2": {"x": 3, "y": 4},
    }


@has_312_features
def test_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(
        left=Tree(
            left=1,
            right=Tree(left=Tree(left=2, right=3), right=Tree(left=4, right=5)),
        ),
        right=Tree(left=Tree(left=6, right=7), right=8),
    )
    assert serialize(Tree[int], tree) == {
        "left": {
            "left": 1,
            "right": {
                "left": {"left": 2, "right": 3},
                "right": {"left": 4, "right": 5},
            },
        },
        "right": {"left": {"left": 6, "right": 7}, "right": 8},
    }


def test_serialize_set():
    nums = {1, 2, 3, 4}
    assert serialize(set[int], nums) == [1, 2, 3, 4]


def test_serialize_frozenset():
    nums = frozenset({1, 2, 3, 4})
    assert serialize(frozenset[int], nums) == [1, 2, 3, 4]


def test_serialize_dict_subclass():
    data = DotDict(apple=3, banana=7)
    assert serialize(DotDict[str, int], data) == {"apple": 3, "banana": 7}


class Special(Medley):
    def serialize(self, typ: type[int], value: int, ctx: Context):
        return value * 10

    def serialize(self, typ: type[int], value: str, ctx: Context):
        return value * 2


def test_override():
    ss = (Serieux + Special)()
    assert ss.serialize(int, 3) == 30
    assert ss.serialize(int, "quack") == "quackquack"
    assert ss.serialize(list[int], [1, 2, 3]) == [10, 20, 30]
    assert ss.serialize(list[int], [1, "2", 3]) == [10, "22", 30]
    assert ss.serialize(Point, Point(8, 9)) == {"x": 80, "y": 90}
    assert ss.serialize(3) == 30


class quirkint(int):
    pass


class Quirky(Medley):
    def serialize(self, typ: type[int], value: quirkint, ctx: Context):
        return value * 10


def test_override_quirkint():
    ss = (Serieux + Quirky)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, quirkint(3)) == 30
    assert ss.serialize(Point, Point(8, 9)) == {"x": 8, "y": 9}
    assert ss.serialize(Point, Point(quirkint(8), 9)) == {"x": 80, "y": 9}


class ExtraWeight(Context):
    weight: int


class WeightedImpl(Medley):
    def serialize(self, typ: type[int], value: int, ctx: ExtraWeight):
        return value + ctx.weight


def test_override_state():
    ss = (Serieux + WeightedImpl)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, 3, ExtraWeight(10)) == 13
    assert ss.serialize(Point, Point(7, 8)) == {"x": 7, "y": 8}
    assert ss.serialize(Point, Point(7, 8), ExtraWeight(10)) == {"x": 17, "y": 18}


def test_serialize_enum():
    assert serialize(Color, Color.RED) == "red"
    assert serialize(list[Color], [Color.GREEN, Color.BLUE, Color.GREEN]) == [
        "green",
        "blue",
        "green",
    ]


def test_serialize_enum_int():
    assert serialize(Level, Level.MED) == 1
    assert serialize(list[Level], [Level.HI, Level.LO, Level.HI]) == [2, 0, 2]


def test_serialize_literal_enum():
    assert serialize(Literal["red", "green", "blue"], "red") == "red"
    with pytest.raises(
        ValidationError,
        match=r"'yellow' is not a valid option for typing.Literal\['red', 'green', 'blue'\]",
    ):
        serialize(Literal["red", "green", "blue"], "yellow")


def test_serialize_literal_int_enum():
    assert serialize(Literal[1, 2, 7], 2) == 2
    with pytest.raises(
        ValidationError,
        match=r"'3' is not a valid option for typing.Literal\[1, 2, 7\]",
    ):
        serialize(Literal[1, 2, 7], 3)


def test_serialize_literal_mixed_enum():
    lit = Literal[1, True, "quack"]
    assert serialize(lit, 1) == 1
    assert serialize(lit, True) is True
    assert serialize(lit, "quack") == "quack"
    with pytest.raises(ValidationError):
        serialize(lit, 3)
    with pytest.raises(ValidationError):
        serialize(lit, "boop")
    with pytest.raises(ValidationError):
        serialize(lit, False)


def test_serialize_ignore(datapath):
    pth = datapath / "job.yaml"
    f = File(pth)
    assert serialize(File, f) == {"path": str(pth)}


###############
# Error tests #
###############


def test_error_basic():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(int, "oh no")


def test_error_dataclass():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(Point, Point(x=1, y="oops"), Trail())


@has_312_features
def test_error_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(Tree("a", 2), "b")

    with pytest.raises(ValidationError, match=r"At path \.left\.right"):
        serialize(Tree[str], tree, Trail())


def test_error_serialize_list():
    li = [0, 1, 2, 3, "oops", 5, 6]

    with pytest.raises(ValidationError, match=r"At path .4"):
        serialize(list[int], li, Trail())


def test_error_serialize_list_of_lists():
    li = [[0, 1], [2, 3, "oops", 5, 6]]

    with pytest.raises(ValidationError, match=r"At path .1.2"):
        serialize(list[list[int]], li, Trail())


def test_dump_no_dest():
    pt = Point(1, 2)
    assert serialize(Point, pt) == dump(Point, pt)


def test_dump(tmp_path):
    dest = tmp_path / "point.yaml"
    pt = Point(1, 2)
    dump(Point, pt, dest=dest)
    assert load(Point, dest) == pt


def test_dump_str_with_format():
    pt = Point(1, 2)
    s = dump(Point, pt, format="yaml")
    assert s == "x: 1\ny: 2\n"


def test_serialize_recursive_type():
    data = DIDHolder(
        did={
            "a": 1,
            "b": {
                "c": 2,
                "d": 3,
            },
        }
    )
    ser = serialize(DIDHolder, data)
    assert ser["did"] == data.did


def test_serialize_recursive_type_2():
    def p(x):
        return Point(x, x)

    data = LTHolder(lt=[p(1), [p(2), [p(3)], p(4)]])
    ser = serialize(LTHolder, data)
    assert ser["lt"] == [
        {"x": 1, "y": 1},
        [{"x": 2, "y": 2}, [{"x": 3, "y": 3}], {"x": 4, "y": 4}],
    ]


def test_error_serialize_list_modelizable():
    th = Thingies(["hello", "world"])
    assert serialize(Thingies, th) == ["hello", "world"]


def test_serialize_defaults_with_omit_defaults_context():
    # Test with default values - serialize normally includes all fields
    obj_with_defaults = Defaults("Alice")
    assert serialize(Defaults, obj_with_defaults) == {
        "name": "Alice",
        "aliases": [],
        "cool": False,
    }

    # Test with OmitDefaults context - should omit fields that have default values
    assert serialize(Defaults, obj_with_defaults, OmitDefaults()) == {
        "name": "Alice",
    }

    # Test with non-default values - should serialize normally
    obj_with_custom_values = Defaults("Bob", aliases=["Robert", "Bobby"], cool=True)
    assert serialize(Defaults, obj_with_custom_values) == {
        "name": "Bob",
        "aliases": ["Robert", "Bobby"],
        "cool": True,
    }

    # Test with OmitDefaults context but non-default values - should include all fields
    assert serialize(Defaults, obj_with_custom_values, OmitDefaults()) == {
        "name": "Bob",
        "aliases": ["Robert", "Bobby"],
        "cool": True,
    }

    # Test partial default values with OmitDefaults
    obj_partial = Defaults("Charlie", cool=True)
    assert serialize(Defaults, obj_partial, OmitDefaults()) == {
        "name": "Charlie",
        "cool": True,
    }


def test_serialize_modify_context():
    alice = Defaults("Alice")
    assert serialize(Defaults @ ModifyContext(OmitDefaults()), alice) == {
        "name": "Alice",
    }
    assert serialize(Defaults @ ModifyContext(sub=OmitDefaults()), alice, OmitDefaults()) == {
        "name": "Alice",
        "aliases": [],
        "cool": False,
    }



from dataclasses import dataclass

from pydantic import TypeAdapter

from .base import Adapter


@dataclass
class PydanticAdapter(Adapter):
    def serializer_for_type(self, t):
        return TypeAdapter(t).serializer.to_python

    def json_for_type(self, t):
        return TypeAdapter(t).serializer.to_json

    def deserializer_for_type(self, t):
        return TypeAdapter(t).validator.validate_python



from dataclasses import replace
from typing import Any

from ovld import Medley, call_next, ovld, recurse
from ovld.types import HasMethod
from ovld.utils import ResolutionError

from ..ctx import Context
from ..model import Model, model
from ..priority import STD4

PRIO = STD4.next()


class UserMethods(Medley):
    @ovld(priority=PRIO)
    def deserialize(self, t: type[HasMethod["serieux_deserialize"]], obj: Any, ctx: Context):  # noqa: F821
        def cn(t, obj, ctx, *, from_top=False):
            return recurse(t, obj, ctx) if from_top else call_next(t, obj, ctx)

        cn.serieux = self

        try:
            # NOTE: we want to preserve the type arguments, but if f is a classmethod,
            # C[T].f == C.f, which will receive C as its first argument. Here we simply
            # make sure that f will indeed receive C[T].
            return t.serieux_deserialize.__func__(t, obj, ctx, cn)
        except ResolutionError:
            # If t implements serieux_deserialize with ovld and no method matches, it will
            # throw a ResolutionError and we simply resume our search down the stack.
            return call_next(t, obj, ctx)

    @ovld(priority=PRIO)
    def serialize(self, t: type[HasMethod["serieux_serialize"]], obj: Any, ctx: Context):  # noqa: F821
        def cn(t, obj, ctx, *, from_top=False):
            return recurse(t, obj, ctx) if from_top else call_next(t, obj, ctx)

        cn.serieux = self

        try:
            return t.serieux_serialize.__func__(t, obj, ctx, cn)
        except ResolutionError:
            return call_next(t, obj, ctx)

    @ovld(priority=PRIO)
    def schema(self, t: type[HasMethod["serieux_schema"]], ctx: Context):  # noqa: F821
        def cn(t, ctx, *, from_top=False):
            return recurse(t, ctx) if from_top else call_next(t, ctx)

        cn.serieux = self

        try:
            return t.serieux_schema.__func__(t, ctx, cn)
        except ResolutionError:  # pragma: no cover
            return call_next(t, ctx)


@model.register(priority=1)
def _(t: type[HasMethod["serieux_model"]]):  # noqa: F821
    def cn(t, *, from_top=False):  # pragma: no cover
        return recurse(t) if from_top else call_next(t)

    return t.serieux_model(cn)


@model.register(priority=2)
def _(t: type[HasMethod["serieux_to_string"]] | type[HasMethod["serieux_from_string"]]):  # noqa: F821
    m = call_next(t)
    if not m:
        m = Model(t, fields=None)
    if hasattr(t, "serieux_to_string"):
        m = replace(m, to_string=t.serieux_to_string)
    if hasattr(t, "serieux_from_string"):
        m = replace(m, from_string=t.serieux_from_string)
    return m



from abc import ABC, abstractmethod


class Adapter(ABC):
    """Base class for all serialization/deserialization adapters."""

    @abstractmethod
    def serializer_for_type(self, t):
        """Return a serializer function for the given type."""
        pass

    @abstractmethod
    def json_for_type(self, t):
        """Return a JSON serializer function for the given type."""
        pass

    @abstractmethod
    def deserializer_for_type(self, t):
        """Return a deserializer function for the given type."""
        pass



from serieux import Serieux
from serieux.features.fromfile import IncludeFile
from serieux.features.interpol import Environment, Interpolation
from tests.definitions import Character

deserialize = (Serieux + IncludeFile + Interpolation)().deserialize


def test_interpol_include(datapath):
    """Test interpolation in $include, e.g. $include: ${env:XYZ}"""
    ca = deserialize(
        Character,
        datapath / "character-dyn.yaml",
        Environment(environ={"BACKSTORY": "jimbo.txt"}),
    )
    cb = deserialize(
        Character,
        datapath / "character-dyn.yaml",
        Environment(environ={"BACKSTORY": "jumbo.txt"}),
    )
    assert ca.backstory != cb.backstory
    assert ca.backstory == (datapath / "jimbo.txt").read_text()
    assert cb.backstory == (datapath / "jumbo.txt").read_text()



from __future__ import annotations

import json
from pathlib import Path

import pytest

from serieux import IncludeFile, Serieux, serialize

from .matrix import Matrix

here = Path(__file__).parent
deserialize = (Serieux + IncludeFile)().deserialize


def bench(path):
    matrix = deserialize(Matrix, path)
    return pytest.mark.parametrize("case", matrix.generate_cases())


@bench(here / "matrix.yaml")
def test_serialize(case, benchmark):
    case.xfail_guard()
    data_ser = serialize(case.data)
    fn = case.adapter.serializer_for_type(type(case.data))
    result = benchmark(fn, case.data)
    assert result == data_ser


@bench(here / "matrix-json.yaml")
def test_json(case, benchmark):
    case.xfail_guard()
    fn = case.adapter.json_for_type(type(case.data))
    result = benchmark(fn, case.data)
    assert json.loads(result) == serialize(type(case.data), case.data)


@bench(here / "matrix.yaml")
def test_deserialize(case, benchmark):
    case.xfail_guard()
    data_ser = serialize(case.data)
    fn = case.adapter.deserializer_for_type(type(case.data))
    result = benchmark(fn, data_ser)
    assert result == case.data



import copy
import math

from serieux.features.comment import CommentProxy
from serieux.proxy import LazyProxy

from .definitions import Point


def test_lazy_proxy():
    lazy_value = LazyProxy(lambda: 42)

    assert lazy_value
    assert lazy_value == 42
    assert str(lazy_value) == "42"
    assert repr(lazy_value) == "42"


def test_lazy_proxy_class_override():
    """__class__ returns _type when set."""
    lazy_with_type = LazyProxy(lambda: 42, type=Point)
    assert lazy_with_type.__class__ is Point


def test_lazy_setattr():
    pt = Point(1, 2)
    lpt = LazyProxy(lambda: pt)
    lpt.x = 333
    assert lpt.x == 333
    assert pt.x == 333


def test_lazy_arithmetic():
    lazy_a = LazyProxy(lambda: 10)
    lazy_b = LazyProxy(lambda: 5)
    lazy_c = LazyProxy(lambda: -3)

    assert lazy_a + lazy_b == 15
    assert lazy_a - lazy_b == 5
    assert lazy_a * lazy_b == 50
    assert lazy_a / lazy_b == 2.0
    assert lazy_a // lazy_b == 2
    assert lazy_a % lazy_b == 0
    assert lazy_a**lazy_b == 100000
    assert abs(lazy_c) == 3
    assert -lazy_c == 3
    assert +lazy_c == -3


def test_lazy_comparisons():
    lazy_value = LazyProxy(lambda: 42)
    assert lazy_value == 42
    assert lazy_value != 43
    assert lazy_value < 100
    assert lazy_value <= 42
    assert lazy_value > 0
    assert lazy_value >= 42


def test_lazy_hash():
    lazy_val = LazyProxy(lambda: 42)
    assert hash(lazy_val) == hash(42)
    assert lazy_val in {42}
    assert {lazy_val: "ok"}[42] == "ok"


def test_lazy_misc():
    laz = LazyProxy(lambda: 37.5)
    assert f"{laz:.3f}" == "37.500"


def test_lazy_list():
    lazy_list = LazyProxy(lambda: [1, 2, 3])

    assert len(lazy_list) == 3
    assert lazy_list[0] == 1
    assert list(lazy_list) == [1, 2, 3]
    assert 2 in lazy_list


# --- Container / sequence ---


def test_lazy_delitem():
    lst = [0, 1, 2, 3]
    lazy_lst = LazyProxy(lambda: lst)
    del lazy_lst[1]
    assert lst == [0, 2, 3]
    assert lazy_lst[0] == 0
    assert lazy_lst[1] == 2


def test_lazy_reversed():
    lazy_list = LazyProxy(lambda: [1, 2, 3])
    assert list(reversed(lazy_list)) == [3, 2, 1]


def test_lazy_index():
    lazy_int = LazyProxy(lambda: 4)
    lst = [10, 20, 30, 40, 50]
    assert lst[lazy_int] == 50


def test_lazy_setitem():
    lst = [0, 1, 2]
    lazy_lst = LazyProxy(lambda: lst)
    lazy_lst[1] = 99
    assert lazy_lst[1] == 99
    assert lst[1] == 99


# --- Context manager ---


def test_lazy_context_manager():
    class Ctx:
        def __enter__(self):
            self.entered = True
            return self

        def __exit__(self, *args):
            self.exited = True
            return False

    ctx = Ctx()
    lazy_ctx = LazyProxy(lambda: ctx)
    with lazy_ctx as c:
        assert c is ctx
        assert ctx.entered
    assert ctx.exited


# --- Attribute deletion ---


def test_lazy_delattr():
    pt = Point(1, 2)
    pt.extra = "temp"
    lpt = LazyProxy(lambda: pt)
    del lpt.extra
    assert not hasattr(pt, "extra")


def test_lazy_delattr_special_attribute():
    """Deleting __special_attributes__ goes through object.__delattr__."""
    proxy = CommentProxy(42, "note")
    assert proxy._ == "note"
    del proxy._
    assert not hasattr(proxy, "_")


# --- Numeric conversion ---


def test_lazy_numeric_conversion():
    lazy_int = LazyProxy(lambda: 42)
    lazy_float = LazyProxy(lambda: 3.7)
    lazy_complex = LazyProxy(lambda: 1 + 2j)

    assert int(lazy_int) == 42
    assert float(lazy_float) == 3.7
    assert complex(lazy_complex) == 1 + 2j


def test_lazy_round_trunc_floor_ceil():
    laz = LazyProxy(lambda: 3.7)
    assert round(laz) == 4
    assert round(laz, 1) == 3.7
    # Use math.* to invoke proxy's __trunc__/__floor__/__ceil__ (not _obj's via getattr)
    assert math.trunc(laz) == 3
    assert math.floor(laz) == 3
    assert math.ceil(laz) == 4


def test_lazy_bytes():
    lazy_bytes = LazyProxy(lambda: b"hello")
    assert bytes(lazy_bytes) == b"hello"


# --- Extended arithmetic ---


def test_lazy_divmod():
    lazy_a = LazyProxy(lambda: 17)
    lazy_b = LazyProxy(lambda: 5)
    assert divmod(lazy_a, lazy_b) == (3, 2)
    assert divmod(17, lazy_b) == (3, 2)


def test_lazy_matmul():
    # Use a simple class since list doesn't support @
    class Vec:
        def __init__(self, x, y):
            self.x, self.y = x, y

        def __matmul__(self, other):
            return self.x * other.x + self.y * other.y

        def __rmatmul__(self, other):
            return other.x * self.x + other.y * self.y

    class LeftOnly:
        """Returns NotImplemented so proxy.__rmatmul__ is used."""

        def __init__(self, x, y):
            self.x, self.y = x, y

        def __matmul__(self, other):
            return NotImplemented

    lazy_a = LazyProxy(lambda: Vec(1, 2))
    lazy_b = LazyProxy(lambda: Vec(3, 4))
    assert (lazy_a @ lazy_b) == 11  # 1*3 + 2*4
    assert (Vec(5, 6) @ lazy_b) == 39  # 5*3 + 6*4
    assert (LeftOnly(1, 2) @ lazy_b) == 11  # triggers __rmatmul__


# --- Bitwise ---


def test_lazy_bitwise():
    lazy_a = LazyProxy(lambda: 0b1010)  # 10
    lazy_b = LazyProxy(lambda: 0b0011)  # 3

    assert (lazy_a & lazy_b) == 2
    assert (lazy_a | lazy_b) == 11
    assert (lazy_a ^ lazy_b) == 9
    assert (lazy_a << 1) == 20
    assert (lazy_a >> 1) == 5
    assert (~lazy_a) == -11


def test_lazy_bitwise_reversed():
    lazy_val = LazyProxy(lambda: 4)
    assert (2 << lazy_val) == 32
    assert (64 >> lazy_val) == 4
    assert (0b1111 & lazy_val) == 4
    assert (0b0001 | lazy_val) == 5
    assert (0b0111 ^ lazy_val) == 3


# --- Copy ---


def test_lazy_copy():
    lst = [1, 2, 3]
    lazy_lst = LazyProxy(lambda: lst)
    copied = copy.copy(lazy_lst)
    assert copied == [1, 2, 3]
    assert copied is not lst


def test_lazy_deepcopy():
    nested = [[1, 2], [3, 4]]
    lazy_nested = LazyProxy(lambda: nested)
    deep_copied = copy.deepcopy(lazy_nested)
    assert deep_copied == [[1, 2], [3, 4]]
    assert deep_copied is not nested
    assert deep_copied[0] is not nested[0]


def test_lazy_deepcopy_explicit():
    """Explicit __deepcopy__ call to cover the memo parameter path."""
    lazy_val = LazyProxy(lambda: {"a": 1})
    memo = {}
    result = type(lazy_val).__deepcopy__(lazy_val, memo)
    assert result == {"a": 1}
    assert result is not lazy_val._obj


# --- Callable ---


def test_lazy_call():
    def greet(name):
        return f"Hello, {name}!"

    lazy_greet = LazyProxy(lambda: greet)
    assert lazy_greet("World") == "Hello, World!"


def test_lazy_proxy_deadlock():
    """Recursive evaluation during computation raises."""

    def recurse():
        return proxy._obj

    proxy = LazyProxy(recurse)
    try:
        _ = proxy._obj
    except Exception as e:
        assert "Deadlock" in str(e)



from dataclasses import dataclass

import orjson as json
from apischema import deserialize as apischema_deserialize, serialize as apischema_serialize

from .base import Adapter


@dataclass
class ApischemaAdapter(Adapter):
    def serializer_for_type(self, t):
        return lambda x: apischema_serialize(t, x, check_type=False)

    def json_for_type(self, t):
        return lambda x: json.dumps(apischema_serialize(t, x, check_type=False))

    def deserializer_for_type(self, t):
        return lambda x: apischema_deserialize(t, x)



from collections import defaultdict

import pytest


@pytest.hookimpl(hookwrapper=True)
def pytest_benchmark_group_stats(config, benchmarks, group_by):
    outcome = yield
    results = defaultdict(list)
    for bench in benchmarks:
        case = bench["params"]["case"]
        an = case.adapter_name
        group = bench["name"].replace(an, "").replace(",]", "]").replace("[]", "")
        results[group].append(bench)
    outcome.force_result(results.items())



import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Generic, TypeVar, get_args, get_origin

from ovld import Medley, ovld

from ..ctx import Context, WorkingDirectory
from ..instructions import BaseInstruction, strip
from ..priority import HI1
from ..proxy import ProxyBase
from ..tell import tells
from .partial import Partial


@dataclass(frozen=True)
class DefaultFactory(BaseInstruction):
    factory: Callable


T = TypeVar("T")
MISSING = object()


class FileBacked(Generic[T]):
    path: Path
    value_type: type
    serieux: object
    context: Context
    timestamp: float = None
    refresh: bool = False
    default_factory: Callable = None
    value: T = None

    def __init__(
        self,
        path: Path,
        value_type: type,
        serieux: object,
        context: Context,
        refresh: bool = False,
        default_factory: Callable = None,
    ):
        self.path = path
        value_type = Partial.strip(value_type)
        self.value_type = value_type
        self.serieux = serieux
        self.context = context
        self.refresh = refresh
        stripped = strip(value_type)
        self.default_factory = default_factory or get_origin(stripped) or stripped
        self._value = None
        self.timestamp = None
        self.load()

    @property
    def value(self):
        if self.refresh and self.path.exists():
            file_mtime = self.path.stat().st_mtime
            if file_mtime > self.timestamp:
                self.load()
        return self._value

    def load(self):
        if self.path.exists():
            self._value = self.serieux.deserialize(self.value_type, self.path, self.context)
            self.timestamp = self.path.stat().st_mtime
        elif self.default_factory:
            self._value = self.default_factory()
            self.timestamp = time.time()
        else:  # pragma: no cover
            raise FileNotFoundError(self.path)

    def save(self, new_value=MISSING):
        if new_value is not MISSING:
            self._value = new_value
        self.serieux.dump(self.value_type, self._value, self.context, dest=self.path)
        self.timestamp = time.time()

    def __str__(self):
        return f"{self._value}@{self.path}"

    __repr__ = __str__


@dataclass(frozen=True)
class FileBackedOptions(BaseInstruction):
    default_factory: Callable | None = None
    refresh: bool = False


@dataclass(frozen=True)
class FileProxy(FileBackedOptions):
    pass


class FileBackedProxy(ProxyBase):
    __special_attributes__ = {
        *ProxyBase.__special_attributes__,
        "_wrapper",
        "_path",
        "load",
        "save",
    }

    def __init__(self, path, value_type, *args, **kwargs):
        self._wrapper = FileBacked(path, value_type, *args, **kwargs)
        self._type = value_type

    @property
    def _obj(self):
        return self._wrapper.value

    @property
    def _path(self):
        return self._wrapper.path

    def load(self):
        return self._wrapper.load()

    def save(self, new_value=MISSING):
        return self._wrapper.save(new_value)

    def __str__(self):
        return str(self._wrapper)

    __repr__ = __str__


PRIO = HI1.next()


def _to_path(obj, ctx):
    if isinstance(ctx, WorkingDirectory):  # pragma: no cover
        return ctx.directory / obj
    else:
        return Path(obj)


class FileBackedFeature(Medley):
    @ovld(priority=PRIO)
    def serialize(self, t: type[FileBacked], obj: FileBacked, ctx):
        return str(obj.path)

    @ovld(priority=PRIO)
    def serialize(self, t: type[Any @ FileProxy], obj: FileBackedProxy, ctx: Context):
        return str(obj._wrapper.path)

    @ovld(priority=PRIO)
    def deserialize(
        self, t: type[FileBacked] | type[FileBacked @ FileBackedOptions], obj: Path | str, ctx
    ):
        cls, fopt = FileBackedOptions.decompose(t)
        (vt,) = get_args(strip(cls))
        extra = vars(fopt) if fopt else {}
        return cls(
            path=_to_path(obj, ctx),
            value_type=vt,
            serieux=self,
            context=ctx,
            **extra,
        )

    @ovld(priority=PRIO)
    def deserialize(self, t: type[Any @ FileProxy], obj: str | Path, ctx: Context):
        # Merging doesn't make sense here so Partial will just cause problems
        t = Partial.strip(t)
        value_type, fb = FileProxy.decompose(t)
        return FileBackedProxy(
            _to_path(obj, ctx),
            value_type,
            self,
            ctx,
            default_factory=fb.default_factory,
            refresh=fb.refresh,
        )

    @ovld(priority=PRIO)
    def schema(self, t: type[FileBacked], ctx):
        return {"type": "string"}

    @ovld(priority=PRIO)
    def schema(self, t: type[Any @ FileProxy], ctx: Context):
        return {"type": "string"}


@tells.register(priority=HI1.next())
def _(expected: type[Any @ FileProxy], given: type[str]):
    return set()


@tells.register(priority=HI1.next())
def _(expected: type[FileBacked], given: type[str]):
    return set()



from dataclasses import dataclass

import orjson as json
from adaptix import Retort

from .base import Adapter

_retort = Retort()


@dataclass
class AdaptixAdapter(Adapter):
    def serializer_for_type(self, t):
        return _retort.get_dumper(t)

    def json_for_type(self, t):
        dump = _retort.get_dumper(t)
        return lambda x: json.dumps(dump(x))

    def deserializer_for_type(self, t):
        return _retort.get_loader(t)



from dataclasses import dataclass

from serde.de import from_dict as serde_from_dict
from serde.json import to_json as serde_to_json
from serde.se import to_dict as serde_to_dict

from .base import Adapter


@dataclass
class SerdeAdapter(Adapter):
    def serializer_for_type(self, t):
        return serde_to_dict

    def json_for_type(self, t):
        return serde_to_json

    def deserializer_for_type(self, t):
        return lambda x: serde_from_dict(t, x)



import re
import sys
from dataclasses import dataclass, field

from ovld import ovld
from ovld.medley import ABSENT

from .ctx import Context, Location, Trail, locate
from .formats import FileSource
from .utils import clsstring


@dataclass
class ContextInformation:
    func: str = None
    path: list[str] = field(default_factory=list)
    locs: list[Location] = field(default_factory=list)
    ctx: Context = None

    @property
    def trail_string(self):
        return "".join([f".{field}" for field in self.path]) if self.path else "(at root)"


def _color(code, text):
    return f"\u001b[1m\u001b[{code}m{text}\u001b[0m"


@ovld
def find_link(o1: dict, o2):
    for k, v in o1.items():
        if v is o2:
            return k
    return None


@ovld
def find_link(o1: list, o2):
    for i, v in enumerate(o1):
        if v is o2:
            return str(i)
    return None


@ovld
def find_link(o1, o2):
    try:
        v1 = vars(o1)
    except TypeError:
        return None
    return find_link(v1, o2)


def extract_information(ctx=None, frame=None):
    frame = frame or sys._getframe(1)
    ci = ContextInformation(ctx=ctx)

    def _fast_path():
        ci.func = ci.func or "<serieux>"
        ci.path[:0] = ci.ctx.trail
        if loc := locate(ci.ctx):
            ci.locs.insert(0, loc)
        return ci

    if isinstance(ci.ctx, Trail):
        return _fast_path()

    above = None
    while frame:
        lcls = frame.f_locals
        ci.ctx = lcls.get("ctx", ci.ctx)
        if isinstance(ci.ctx, Trail):
            return _fast_path()
        elif ci.ctx and (m := re.match(r"^((?:de)?serialize|schema)\[", frame.f_code.co_name)):
            ci.func = m.groups()[0]
            t1, obj1 = lcls.get("t", ABSENT), lcls.get("obj", ABSENT)
            obj2 = None
            if above is not None:
                _, obj2 = above
                if lnk := find_link(obj1, obj2):
                    ci.path.insert(0, lnk)
            if obj1 is not obj2 and isinstance(obj1, FileSource):
                try:
                    fp = obj1.field.split(".") if obj1.field else []
                    loc = obj1.format.locate(obj1.path, [*fp, *ci.path])
                    ci.locs.insert(0, loc)
                except FileNotFoundError:
                    pass
            above = t1, obj1
        frame = frame.f_back
    return ci


def display_context_information(
    template="An error happened in serieux.{func} at location {trail}",
    *,
    ctx=None,
    exc=None,
    frame=None,
    show_source=True,
    file=sys.stderr,
    message="ERROR",
    **kwargs,
):
    ci = None
    if exc is not None:
        if isinstance(exc, SerieuxError) and exc.ctx:
            ctx = exc.ctx
            ci = exc.info
        elif isinstance(exc, SerieuxExceptionGroup):  # pragma: no cover
            display(exc)
            return
        else:
            tb = exc.__traceback__
            while tb:
                frame = tb.tb_frame
                tb = tb.tb_next
    ci = ci or extract_information(ctx, frame)
    if ci.func is None:  # pragma: no cover
        return
    print(
        template.format(trail=_color(33, ci.trail_string), func=ci.func, message=message),
        file=file,
    )
    if show_source and ci.locs:
        for location in ci.locs:
            display_location(location, file=file, **kwargs)


def display_location(location, source_context=1, indent=0, ellipsis_cutoff=3, file=sys.stderr):
    width = 3 + indent
    (l1, c1), (l2, c2) = location.linecols
    if c2 == 0:
        l2 -= 1
        c2 = 10_000_000_000_000
    lines = location.whole_text.split("\n")
    start = l1 - source_context
    while start < 0 or not lines[start].strip():
        start += 1
    end = l2 + source_context
    while end >= len(lines) or not lines[end].strip():
        end -= 1

    print(f"{'':{indent}}@ {location.source.absolute()}:{l1 + 1}", file=file)
    for li in range(start, end + 1):
        line = lines[li]
        if li == l2 and not line.strip():  # pragma: no cover
            break
        if li == l1 + ellipsis_cutoff and li < l2:
            print(f"{'':{width}}  ...", file=file)
            continue
        elif li > l1 + ellipsis_cutoff and li < l2:
            continue

        hls = hle = 0
        if li == l1:
            hls = c1
        if li >= l1 and li < l2:
            hle = len(line)
        if li == l2:
            hle = c2

        if hls or hle:
            line = line[:hls] + _color(31, line[hls:hle]) + line[hle:]

        print(f"{li + 1:{width}}: {line}", file=file)


def merge_errors(*errors):
    collected = []
    for err in errors:
        if isinstance(err, ExceptionGroup):
            collected.extend(err.exceptions)
        elif isinstance(err, Exception):
            collected.append(err)
    match collected:
        case ():  # pragma: no cover
            return None
        case (err,):
            return err
        case many:
            return SerieuxExceptionGroup("Some errors occurred", many)


class BaseSerieuxError(Exception):
    pass


class SerieuxError(BaseSerieuxError):
    def __init__(self, message=None, *, ctx=None):
        super().__init__(message)
        self.info = extract_information(ctx)
        self.ctx = self.info.ctx

    @property
    def message(self):
        return self.args[0]

    def __str__(self):
        if location := locate(self.ctx):
            (l1, c1), (l2, c2) = location.linecols
            lc = f"{l1}:{c1}-{l2}:{c2}" if l1 != l2 else f"{l1}:{c1}-{c2}"
            return f"{location.source}:{lc} -- {self.message}"
        else:
            return f"At path {self.info.trail_string}: {self.message}"


class NotGivenError(SerieuxError):
    pass


class MissingFieldError(SerieuxError):
    def __init__(self, t, field_name, *, ctx=None):
        self.t = t
        self.field_name = field_name
        super().__init__(
            message=f"Missing required field '{self.field_name}' for type `{clsstring(t)}`",
            ctx=ctx,
        )


class UnrecognizedFieldError(SerieuxError):
    def __init__(self, t, expected, found, *, ctx=None):
        self.t = t
        self.expected = set(expected)
        self.found = set(found)
        self.unrecognized = found - expected
        super().__init__(
            message=f"Extra unrecognized fields were found for type `{clsstring(t)}`: {self.unrecognized}",
            ctx=ctx,
        )


class WrappedSerieuxError(SerieuxError):
    def __init__(self, message=None, *, exc=None, ctx=None):
        if message is None:
            message = f"{type(exc).__name__}: {exc}"
        super().__init__(message=message, ctx=ctx)
        self.exc = exc


class ValidationError(WrappedSerieuxError):
    pass


class SchemaError(WrappedSerieuxError):
    pass


class SerieuxExceptionGroup(BaseSerieuxError, ExceptionGroup):
    def derive(self, excs):  # pragma: no cover
        return SerieuxExceptionGroup(self.message, excs)


def display(exc, file=sys.stderr):
    if isinstance(exc, SerieuxExceptionGroup):
        for i, subexc in enumerate(exc.exceptions):
            print(f"[#{i}] ", end="", file=file)
            display(subexc, file)
    else:
        if isinstance(exc, SerieuxError):
            msg = exc.message
        else:
            msg = f"{type(exc).__name__}: {exc}"
        display_context_information(
            "{trail}: {message}",
            exc=exc,
            file=file,
            message=msg,
            indent=2,
        )



from dataclasses import dataclass

import orjson as json

from serieux import get_deserializer, get_serializer

from .base import Adapter


@dataclass
class SerieuxAdapter(Adapter):
    def serializer_for_type(self, t):
        return get_serializer(t)

    def json_for_type(self, t):
        func = get_serializer(t)
        return lambda x: json.dumps(func(x))

    def deserializer_for_type(self, t):
        return get_deserializer(t)



from dataclasses import dataclass

import marshmallow_dataclass
import orjson as json

from .base import Adapter


@dataclass
class MarshmallowAdapter(Adapter):
    def serializer_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return schema.dump

    def json_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return lambda x: json.dumps(schema.dump(x))

    def deserializer_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return schema.load



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from numbers import Number
from pathlib import Path

from serieux import Field, Model
from serieux.ctx import Context


@dataclass
class Tree:
    left: Tree | Number
    right: Tree | Number


@dataclass
class Elf:
    name: str
    birthdate: date
    favorite_color: str


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Point3D(Point):
    z: int


# This one is not a dataclass and useful error messages should point that out
class Pointato:
    x: int
    y: int


class Color(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Level(Enum):
    HI = 2
    MED = 1
    LO = 0


@dataclass
class Pig:
    # How pink the pig is
    pinkness: float

    weight: float
    """Weight of the pig, in kilograms"""

    # Is the pig...
    # truly...
    beautiful: bool = True  # ...beautiful?


@dataclass
class Defaults:
    name: str
    aliases: list[str] = field(default_factory=list)
    cool: bool = field(default=False, kw_only=True)


@dataclass
class Player:
    first: str
    last: str
    batting: float


@dataclass
class Team:
    name: str
    players: list[Player]


@dataclass(frozen=True)
class Job:
    # Name of the job
    title: str
    # How much it pays, in dollars
    yearly_pay: float


@dataclass
class Worker:
    name: str
    job: Job = None


DID = dict[str, "int | DID"]


@dataclass
class DIDHolder:
    did: DID


ListTree = list["Point | ListTree"]


@dataclass
class LTHolder:
    lt: ListTree


class DotDict(dict):
    def __getattr__(self, attr):
        return self[attr]


@dataclass
class Character:
    name: str
    age: int
    occupation: str
    backstory: str

    class SerieuxConfig:
        allow_extras = True


@dataclass(kw_only=True)
class Car:
    horsepower: int


@dataclass
class IdentifiedCar(Car):
    id: int


@dataclass
class File:
    path: Path

    # [serieux: ignore]
    fd: object = None

    def __post_init__(self):
        self.fd = open(self.path, "r")


@dataclass
class Thingies:
    things: list[str]

    @classmethod
    def serieux_model(cls, call_next):
        return Model(
            original_type=cls,
            element_field=Field(type=str),
            from_list=cls,
            to_list=lambda x: x.things,
        )


class PrefixContext(Context):
    prefix: str


@dataclass
class ContextSwitched:
    word: str

    @classmethod
    def serieux_serialize(cls, obj, ctx, call_next):
        if isinstance(ctx, PrefixContext):
            return {"word": f"{ctx.prefix}{obj.word}"}
        else:
            return {"word": obj.word}

    @classmethod
    def serieux_deserialize(cls, obj, ctx, call_next):
        word = obj["word"]
        if isinstance(ctx, PrefixContext) and word.startswith(ctx.prefix):
            word = word[len(ctx.prefix) :]
        return cls(word=word)

    @classmethod
    def serieux_schema(cls, ctx, call_next):
        if isinstance(ctx, PrefixContext):
            word_schema = {"type": "string", "pattern": f"^{ctx.prefix}.*"}
        else:
            word_schema = {"type": "string"}
        return {"type": "object", "properties": {"word": word_schema}}



from __future__ import annotations

import itertools
from dataclasses import dataclass, field
from typing import Any

import pytest

from benchmarks.adapters.base import Adapter
from serieux.features.registered import Referenced
from serieux.features.tagset import TaggedSubclass


@dataclass
class Matrix:
    adapters: dict[str, TaggedSubclass[Adapter]]
    data: dict[str, Referenced[Any]]
    xfails: list[list[str]] = field(default_factory=list)

    def generate_cases(self):
        for (dn, d), (an, a) in itertools.product(self.data.items(), self.adapters.items()):
            yield Case(
                adapter=a,
                data=d,
                adapter_name=an,
                data_name=dn,
                xfail=[an, dn] in self.xfails,
            )


@dataclass
class Case:
    adapter: Adapter
    data: Any
    adapter_name: str
    data_name: str
    xfail: bool = False

    def __post_init__(self):
        self.__name__ = f"{self.data_name},{self.adapter_name}"

    def xfail_guard(self):
        if self.xfail:
            pytest.xfail("known failure for this data/interface combination")



from pathlib import Path

from ..ctx import Location
from .abc import FileFormat


class Text(FileFormat):
    def locate(self, f: Path, trail: tuple[str]):
        if trail == ():
            txt = f.read_text()
            lines = txt.splitlines(keepends=True)
            return Location(
                source=f,
                start=0,
                end=len(txt),
                linecols=((0, 0), (len(lines), len(lines[-1]))),
            )

    def patch(self, source, patches):
        for start, end, content in sorted(patches, reverse=True):
            source = source[:start] + content + source[end:]
        return source

    def loads(self, s: str):
        return s

    def dumps(self, data):
        return data



from dataclasses import dataclass

from mashumaro.codecs.basic import BasicDecoder, BasicEncoder
from mashumaro.codecs.orjson import ORJSONEncoder

from .base import Adapter


@dataclass
class MashumaroAdapter(Adapter):
    def serializer_for_type(self, t):
        return BasicEncoder(t).encode

    def json_for_type(self, t):
        return ORJSONEncoder(t).encode

    def deserializer_for_type(self, t):
        return BasicDecoder(t).decode



from dataclasses import dataclass
from typing import Annotated

import pytest

from serieux import Serieux, schema
from serieux.exc import ValidationError
from serieux.features.comment import (
    Comment,
    CommentedObjects,
    CommentProxy,
    CommentRec,
    StripComments,
    comment_field,
)
from serieux.features.tagset import value_field

featured = (Serieux + CommentedObjects)()
serialize = featured.serialize
deserialize = featured.deserialize


@dataclass
class Person:
    name: str
    age: int


def test_commented_proxy():
    obj = [1, 2, 3]
    comment = "This is a comment"
    proxy = CommentProxy(obj, comment)

    # The proxy should behave like the original object
    assert proxy[0] == 1
    assert list(proxy) == [1, 2, 3]
    assert 2 in proxy
    assert str(proxy) == str(obj)
    assert repr(proxy) == repr(obj)

    # The comment should be accessible via the "_" attribute
    assert proxy._ == comment

    # The underlying object should be accessible via _obj
    assert proxy._obj is obj


def test_comment_serialize_int():
    proxy = CommentProxy(42, "The answer")
    serialized = serialize(Comment[int, str], proxy)
    expected = {value_field: 42, comment_field: "The answer"}
    assert serialized == expected


def test_comment_deserialize_int():
    serialized = {value_field: 42, comment_field: "The answer"}
    deserialized = deserialize(Comment[int, str], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert deserialized._obj == 42
    assert deserialized._ == "The answer"


def test_comment_serialize_person():
    person = Person(name="Alice", age=30)
    proxy = CommentProxy(person, "A nice person")
    serialized = serialize(Comment[Person, str], proxy)
    expected = {"name": "Alice", "age": 30, comment_field: "A nice person"}
    assert serialized == expected


def test_comment_deserialize_person():
    serialized = {"name": "Alice", "age": 30, comment_field: "A nice person"}
    deserialized = deserialize(Comment[Person, str], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert isinstance(deserialized._obj, Person)
    assert deserialized._obj == Person(name="Alice", age=30)
    assert deserialized._ == "A nice person"


def test_comment_with_complex_comment_type():
    comment_person = Person("Bob", 25)
    proxy = CommentProxy(100, comment_person)

    serialized = serialize(Comment[int, Person], proxy)
    expected = {value_field: 100, comment_field: {"name": "Bob", "age": 25}}
    assert serialized == expected

    deserialized = deserialize(Comment[int, Person], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert deserialized._obj == 100
    assert deserialized._ == comment_person


def test_comment_without_comment_field():
    data = {"name": "Charlie", "age": 35}
    result = deserialize(Comment[Person, str], data)
    assert type(result) is Person
    assert result.name == "Charlie"
    assert result.age == 35


def test_comment_required_not_commentproxy():
    person = Person("Eve", 28)
    with pytest.raises(ValidationError):
        serialize(Annotated[Person, Comment(str, True)], person)


def test_comment_required_without_comment():
    data = {"name": "Charlie", "age": 35}
    with pytest.raises(ValidationError):
        deserialize(Annotated[Person, Comment(str, True)], data)


def test_comment_required_without_comment_int():
    data = 100
    with pytest.raises(ValidationError):
        deserialize(Annotated[int, Comment(str, True)], data)


def test_comment_serialize_non_proxy():
    person = Person("David", 40)
    serialized = serialize(Comment[Person, str], person)
    expected = {"name": "David", "age": 40}
    assert serialized == expected


def test_comment_serialize_proxy_to_normal_type():
    person = Person("David", CommentProxy(40, "v. young"))
    serialized = serialize(Person, person)
    expected = {"name": "David", "age": 40}
    assert serialized == expected


def test_comment_schema(schematest):
    schematest(
        type=Comment[Person, str],
        value=CommentProxy(Person("Alice", 30), "note"),
    )


def test_comment_schema_required(schematest):
    schematest(
        type=Annotated[Person, Comment(str, required=True)],
        value=CommentProxy(Person("Bob", 25), "required comment"),
    )


def test_comment_schema_primitive(schematest):
    schematest(
        type=Annotated[str, Comment(str, required=True)],
        value=CommentProxy("hello", "world"),
    )


def test_commentrec_serialize():
    person = Person("David", CommentProxy(40, "v. young"))

    serialized = serialize(CommentRec[Person, str], person)
    expected = {"name": "David", "age": {value_field: 40, comment_field: "v. young"}}
    assert serialized == expected

    deserialized = deserialize(CommentRec[Person, str], expected)
    assert deserialized == person
    assert deserialized.age._ == "v. young"


def test_commentrec_list():
    david = Person("David", CommentProxy(40, "v. young"))
    helen = Person("Helen", CommentProxy(80, "v. old"))
    people = [david, helen]

    serialized = serialize(CommentRec[list[Person], str], people)
    expected = [
        {"name": "David", "age": {value_field: 40, comment_field: "v. young"}},
        {"name": "Helen", "age": {value_field: 80, comment_field: "v. old"}},
    ]
    assert serialized == expected

    deserialized = deserialize(CommentRec[list[Person], str], expected)
    assert deserialized == people
    assert deserialized[0].age._ == "v. young"
    assert deserialized[1].age._ == "v. old"


def test_strip_comments_serialize():
    sc = StripComments()
    person = Person("David", CommentProxy(40, "v. young"))
    assert serialize(CommentRec[Person, str], person, sc) == serialize(Person, person, sc)


def test_strip_comments_deserialize():
    sc = StripComments()
    person = {"name": "David", "age": {value_field: 40, comment_field: "v. young"}}
    des = deserialize(CommentRec[Person, str], person, sc)
    assert type(des.age) is int


def test_strip_comments_schema():
    sc = StripComments()
    assert schema(Comment[Person, str], sc).compile() == schema(Person, sc).compile()
    assert schema(CommentRec[Person, str], sc).compile() == schema(Person, sc).compile()



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Tree:
    left: Tree | int
    right: Tree | int


tree = Tree(1, Tree(Tree(2, 3), Tree(4, Tree(5, Tree(6, Tree(7, Tree(8, 9)))))))



from functools import cached_property


class ProxyBase:
    __special_attributes__ = {"_obj", "_computing", "_evaluate", "_type", "__dict__", "_"}

    def __getattribute__(self, name):
        if name in type(self).__special_attributes__:
            return object.__getattribute__(self, name)
        elif name == "__class__":
            return object.__getattribute__(self, "_type") or type(self)
        return getattr(self._obj, name)

    def __setattr__(self, name, value):
        if name in type(self).__special_attributes__:
            return object.__setattr__(self, name, value)
        return setattr(self._obj, name, value)

    def __delattr__(self, name):
        if name in type(self).__special_attributes__:
            return object.__delattr__(self, name)
        return delattr(self._obj, name)

    def __str__(self):
        return str(self._obj)

    def __repr__(self):
        return repr(self._obj)

    def __eq__(self, other):
        return self._obj == other

    def __ne__(self, other):
        return self._obj != other

    def __lt__(self, other):
        return self._obj < other

    def __le__(self, other):
        return self._obj <= other

    def __gt__(self, other):
        return self._obj > other

    def __ge__(self, other):
        return self._obj >= other

    def __hash__(self):
        return hash(self._obj)

    def __len__(self):
        return len(self._obj)

    def __getitem__(self, key):
        return self._obj[key]

    def __setitem__(self, key, value):
        self._obj[key] = value

    def __delitem__(self, key):
        del self._obj[key]

    def __iter__(self):
        return iter(self._obj)

    def __bool__(self):
        return bool(self._obj)

    def __contains__(self, item):
        return item in self._obj

    def __reversed__(self):
        return reversed(self._obj)

    def __index__(self):
        return self._obj.__index__()

    def __enter__(self):
        return self._obj.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._obj.__exit__(exc_type, exc_val, exc_tb)

    def __int__(self):
        return int(self._obj)

    def __float__(self):
        return float(self._obj)

    def __complex__(self):
        return complex(self._obj)

    def __round__(self, ndigits=None):
        return round(self._obj, ndigits) if ndigits is not None else round(self._obj)

    def __trunc__(self):
        return self._obj.__trunc__()

    def __floor__(self):
        return self._obj.__floor__()

    def __ceil__(self):
        return self._obj.__ceil__()

    def __bytes__(self):
        return bytes(self._obj)

    def __add__(self, other):
        return self._obj + other

    def __sub__(self, other):
        return self._obj - other

    def __mul__(self, other):
        return self._obj * other

    def __truediv__(self, other):
        return self._obj / other

    def __floordiv__(self, other):
        return self._obj // other

    def __mod__(self, other):
        return self._obj % other

    def __pow__(self, other):
        return self._obj**other

    def __radd__(self, other):
        return other + self._obj

    def __rsub__(self, other):
        return other - self._obj

    def __rmul__(self, other):
        return other * self._obj

    def __rtruediv__(self, other):
        return other / self._obj

    def __rfloordiv__(self, other):
        return other // self._obj

    def __rmod__(self, other):
        return other % self._obj

    def __rpow__(self, other):
        return other**self._obj

    def __divmod__(self, other):
        return divmod(self._obj, other)

    def __rdivmod__(self, other):
        return divmod(other, self._obj)

    def __matmul__(self, other):
        return self._obj @ other

    def __rmatmul__(self, other):
        return other @ self._obj

    def __lshift__(self, other):
        return self._obj << other

    def __rshift__(self, other):
        return self._obj >> other

    def __and__(self, other):
        return self._obj & other

    def __or__(self, other):
        return self._obj | other

    def __xor__(self, other):
        return self._obj ^ other

    def __rlshift__(self, other):
        return other << self._obj

    def __rrshift__(self, other):
        return other >> self._obj

    def __rand__(self, other):
        return other & self._obj

    def __ror__(self, other):
        return other | self._obj

    def __rxor__(self, other):
        return other ^ self._obj

    def __invert__(self):
        return ~self._obj

    def __neg__(self):
        return -self._obj

    def __pos__(self):
        return +self._obj

    def __abs__(self):
        return abs(self._obj)

    def __format__(self, fmt):
        return format(self._obj, fmt)

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)

    def __copy__(self):
        import copy

        return copy.copy(self._obj)

    def __deepcopy__(self, memo):
        import copy

        return copy.deepcopy(self._obj, memo)


class LazyProxy(ProxyBase):
    def __init__(self, evaluate, type=None):
        self._type = type
        self._evaluate = evaluate
        self._computing = False

    @cached_property
    def _obj(self):
        if self._computing:
            raise Exception("Deadlock: asked for a value during its computation.")
        self._computing = True
        try:
            rval = self._evaluate()
            if isinstance(rval, LazyProxy):  # pragma: no cover
                return rval._obj
        finally:
            self._computing = False
        return rval



import time
from dataclasses import dataclass

import pytest

from serieux import Serieux
from serieux.features.filebacked import (
    FileBacked,
    FileBackedFeature,
    FileBackedOptions,
    FileProxy,
)
from serieux.tell import tells

from ..definitions import Point

srx = (Serieux + FileBackedFeature)()


@dataclass
class Config:
    point: Point @ FileProxy(default_factory=lambda: Point(0, 0))


@dataclass
class ConfigNoDefault:
    point: Point @ FileProxy()


def test_filebacked_deserialize(tmp_path):
    point_file = tmp_path / "point.yaml"

    srx.dump(Point, Point(2, 3), dest=point_file)

    pf = srx.deserialize(FileBacked[Point], str(point_file))

    assert pf.value.x == 2
    assert pf.value.y == 3
    assert pf.path == point_file

    pf.value.x = 15
    pf.value.y = 25

    pf.save()

    reloaded = srx.deserialize(Point, point_file)
    assert reloaded.x == 15
    assert reloaded.y == 25

    pf.save(Point(7, 19))

    reloaded = srx.deserialize(Point, point_file)
    assert reloaded.x == 7
    assert reloaded.y == 19


def test_filebacked_serialize(tmp_path):
    point_file = tmp_path / "point.yaml"
    srx.dump(Point, Point(8, 9), dest=point_file)
    config = srx.deserialize(FileBacked[Point], str(point_file))
    serialized = srx.serialize(FileBacked[Point], config)
    assert serialized == str(point_file)


def test_filebacked_default_factory(tmp_path):
    point_file = tmp_path / "nonexistent.yaml"

    pf = srx.deserialize(
        FileBacked[Point] @ FileBackedOptions(default_factory=lambda: Point(5, 10)),
        str(point_file),
    )

    assert pf.value.x == 5
    assert pf.value.y == 10

    pf.value.x = 3
    pf.value.y = 7
    pf.save()

    assert point_file.exists()
    reloaded = srx.deserialize(Point, point_file)
    assert reloaded.x == 3
    assert reloaded.y == 7


def test_filebacked_proxy_basic(tmp_path):
    point_file = tmp_path / "point.yaml"

    srx.dump(Point, Point(1, 2), dest=point_file)

    config = srx.deserialize(Config, {"point": str(point_file)})

    assert config.point.x == 1
    assert config.point.y == 2
    assert config.point._path == point_file

    config.point.x = 10
    config.point.y = 20

    config.point.save()

    reloaded = srx.deserialize(Point, point_file)
    assert reloaded.x == 10
    assert reloaded.y == 20


def test_filebacked_proxy_default_factory(tmp_path):
    point_file = tmp_path / "nonexistent.yaml"

    config = srx.deserialize(Config, {"point": str(point_file)})

    assert config.point.x == 0
    assert config.point.y == 0

    config.point.x = 3
    config.point.y = 7
    config.point.save()

    assert point_file.exists()
    reloaded = srx.deserialize(Point, point_file)
    assert reloaded.x == 3
    assert reloaded.y == 7


def test_fileproxy_no_default_good(tmp_path):
    point_file = tmp_path / "missing.yaml"

    lp = srx.deserialize(list[Point] @ FileProxy(), str(point_file))
    assert lp == []


def test_fileproxy_no_default_bad(tmp_path):
    point_file = tmp_path / "missing.yaml"

    with pytest.raises(TypeError, match="missing 2 required positional arguments"):
        srx.deserialize(ConfigNoDefault, {"point": str(point_file)})


def test_fileproxy_reload(tmp_path):
    point_file = tmp_path / "point.yaml"

    srx.dump(Point, Point(1, 1), dest=point_file)

    config = srx.deserialize(Config, {"point": str(point_file)})
    assert config.point.x == 1
    assert config.point.y == 1

    srx.dump(Point, Point(99, 99), dest=point_file)

    config.point.load()

    assert config.point.x == 99
    assert config.point.y == 99


def test_fileproxy_serialize(tmp_path):
    point_file = tmp_path / "point.yaml"
    srx.dump(Point, Point(4, 5), dest=point_file)
    config = srx.deserialize(Config, {"point": str(point_file)})
    serialized = srx.serialize(Config, config)
    assert serialized == {"point": str(point_file)}


def test_filebacked_str_repr(tmp_path):
    point_file = tmp_path / "point.yaml"

    srx.dump(Point, Point(7, 8), dest=point_file)

    pf = srx.deserialize(FileBacked[Point], str(point_file))
    str_repr = str(pf)
    assert str_repr == repr(pf)
    assert "Point(x=7, y=8)@" in str_repr
    assert str(point_file) in str_repr


def test_fileproxy_str_repr(tmp_path):
    point_file = tmp_path / "point.yaml"

    srx.dump(Point, Point(7, 8), dest=point_file)

    pf = srx.deserialize(Point @ FileProxy, str(point_file))
    str_repr = str(pf)
    assert str_repr == repr(pf)
    assert "Point(x=7, y=8)@" in str_repr
    assert str(point_file) in str_repr


def test_fileproxy_tells():
    assert tells(expected=Point @ FileProxy(), given=str) == set()


def test_filebacked_tells():
    assert tells(expected=FileBacked[Point], given=str) == set()


def test_filebacked_schema():
    schema = srx.schema(FileBacked[Point]).compile(root=False)
    assert schema == {"type": "string"}


def test_fileproxy_schema():
    schema = srx.schema(Point @ FileProxy).compile(root=False)
    assert schema == {"type": "string"}


def test_filebacked_refresh(tmp_path):
    point_file = tmp_path / "point.yaml"

    # Write initial point
    srx.dump(Point, Point(1, 2), dest=point_file)

    # FileBacked should load initial value
    fb = srx.deserialize(FileBacked[Point], str(point_file))
    fbr = srx.deserialize(FileBacked[Point] @ FileBackedOptions(refresh=True), str(point_file))
    assert fb.value == Point(1, 2)
    assert fbr.value == Point(1, 2)

    # Change the point file after a small delay
    time.sleep(0.01)
    srx.dump(Point, Point(9, 10), dest=point_file)
    assert fb.value == Point(1, 2)
    assert fbr.value == Point(9, 10)

    # Change the point again
    time.sleep(0.01)
    srx.dump(Point, Point(99, -3), dest=point_file)
    assert fb.value == Point(1, 2)
    assert fbr.value == Point(99, -3)


def test_filebacked_refresh_proxy(tmp_path):
    point_file = tmp_path / "point.yaml"

    # Write initial point
    srx.dump(Point, Point(1, 2), dest=point_file)

    # FileBacked should load initial value
    fb = srx.deserialize(Point @ FileProxy(), str(point_file))
    fbr = srx.deserialize(Point @ FileProxy(refresh=True), str(point_file))
    assert fb == Point(1, 2)
    assert fbr == Point(1, 2)

    # Change the point file after a small delay
    time.sleep(0.01)
    srx.dump(Point, Point(9, 10), dest=point_file)
    assert fb == Point(1, 2)
    assert fbr == Point(9, 10)

    # Change the point again
    time.sleep(0.01)
    srx.dump(Point, Point(99, -3), dest=point_file)
    assert fb == Point(1, 2)
    assert fbr == Point(99, -3)
